<?php

declare(strict_types=1);

namespace Drupal\Tests\data_pipelines_elasticsearch\Kernel;

use Drupal\data_pipelines\DatasetPipelinePluginManager;
use Drupal\data_pipelines_elasticsearch\Plugin\DatasetDestination\ElasticSearchDestination;
use Drupal\Tests\data_pipelines\Kernel\DatasetKernelTestBase;
use Drupal\Tests\data_pipelines_elasticsearch\Traits\ElasticsearchTestDestinationTrait;
use Elasticsearch\Client;
use Elasticsearch\Namespaces\IndicesNamespace;
use Prophecy\Argument;
use Prophecy\PhpUnit\ProphecyTrait;
use Psr\Log\NullLogger;

/**
 * Defines a class for testing elastic indexer.
 *
 * @group data_pipelines
 */
class ElasticsearchDestinationDatasetTest extends DatasetKernelTestBase {

  use ProphecyTrait;
  use ElasticsearchTestDestinationTrait;

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'data_pipelines_elasticsearch',
    'data_pipelines_elasticsearch_test',
  ];

  /**
   * Tests elasticsearch destinations.
   */
  public function testElasticsearchDestination() {
    $destination = $this->createElasticsearchDestination();
    $dataset = $this->createTestDataset([
      'destinations' => $destination,
      'pipeline' => 'elasticsearch_test_pipeline_1',
    ]);
    $this->assertCount(1, $dataset->getDestinations());
  }

  /**
   * Tests multiple destinations.
   */
  public function testMultipleDestinations() {
    $destinations[] = $this->createTestMemoryDestination();
    $destinations[] = $this->createTestFileDestination();
    $destinations[] = $this->createElasticsearchDestination();
    $dataset = $this->createTestDataset(['destinations' => $destinations]);
    $this->assertCount(3, $dataset->getDestinations());
  }

  /**
   * Tests elasticsearch data pipeline plugin.
   */
  public function testElasticsearchDataPipeline() {
    $destination = $this->createElasticsearchDestination();
    $dataset = $this->createTestDataset([
      'destinations' => $destination,
      'pipeline' => 'elasticsearch_test_pipeline_1',
    ]);
    $indices = $this->prophesize(IndicesNamespace::class);
    $indices->exists(Argument::any())->willReturn(FALSE);
    $params = [
      'index' => 'foo_' . $dataset->getMachineName(),
      'body' => [
        'settings' => [
          'number_of_shards' => ElasticSearchDestination::INDEX_SHARDS,
          'number_of_replicas' => ElasticSearchDestination::INDEX_REPLICAS,
          'codec' => ElasticSearchDestination::INDEX_CODEC,
        ],
        'mappings' => [
          'properties' => [
            'firstname' => [
              'type' => 'text',
              'index' => FALSE,
            ],
            'lastname' => [
              'type' => 'text',
              'index' => FALSE,
            ],
            'full_name' => [
              'type' => 'text',
            ],
          ],
        ],
      ],
    ];
    $indices->create($params)->shouldBeCalled();
    $client = $this->prophesize(Client::class);
    $client->indices()->willReturn($indices->reveal());

    $client->bulk(Argument::any())->willReturn([]);

    $pipeline_manager = $this->prophesize(DatasetPipelinePluginManager::class);
    $pipeline_manager->hasDefinition(Argument::exact('elasticsearch_test_pipeline_1'))->willReturn(TRUE);
    $pipeline_manager->getDefinition(Argument::exact('elasticsearch_test_pipeline_1'))->willReturn(['mappings' => $params['body']['mappings']]);

    $configuration = [
      'prefix' => 'foo_',
    ];
    $plugin_id = 'foo';
    $plugin_definition = [];
    $logger = new NullLogger();
    $indexer = new ElasticSearchDestination($configuration, $plugin_id, $plugin_definition, $pipeline_manager->reveal(), $logger, $client->reveal());
    $this->assertTrue($indexer->beginProcessing($dataset));
    $this->assertTrue($indexer->processChunk($dataset, iterator_to_array($dataset->getDataIterator())));
  }

  /**
   * Tests cleanup processing.
   */
  public function testCleanup(): void {
    $destination = $this->createElasticsearchDestination();
    $dataset = $this->createTestDataset([
      'destinations' => $destination,
      'pipeline' => 'elasticsearch_test_pipeline_1',
    ]);
    $indices = $this->prophesize(IndicesNamespace::class);
    $indices->exists(Argument::any())->willReturn(TRUE);
    $client = $this->prophesize(Client::class);
    $client->indices()->willReturn($indices->reveal());

    $client->bulk([
      'body' => [
        [
          'delete' => [
            '_index' => 'foo_' . $dataset->getMachineName(),
            '_id' => $dataset->getMachineName() . ':' . 1,
          ],
        ],
        [
          'delete' => [
            '_index' => 'foo_' . $dataset->getMachineName(),
            '_id' => $dataset->getMachineName() . ':' . 2,
          ],
        ],
      ],
    ])->willReturn([]);

    $configuration = [
      'prefix' => 'foo_',
    ];
    $plugin_id = 'foo';
    $plugin_definition = [];
    $logger = new NullLogger();
    $pipeline_manager = $this->prophesize(DatasetPipelinePluginManager::class);
    $indexer = new ElasticSearchDestination($configuration, $plugin_id, $plugin_definition, $pipeline_manager->reveal(), $logger, $client->reveal());
    $this->assertTrue($indexer->processCleanup($dataset, [1, 2]));
  }

}
