<?php

declare(strict_types=1);

namespace Drupal\Tests\data_pipelines_elasticsearch\Kernel;

use Drupal\data_pipelines\Destination\ProcessingOperationEnum;
use Drupal\data_pipelines\Entity\DatasetInterface;
use Drupal\Tests\data_pipelines\Kernel\DatasetKernelTestBase;
use Drupal\Tests\data_pipelines\Traits\DatasetTestProcessingTrait;
use Drupal\Tests\data_pipelines_elasticsearch\Traits\ElasticsearchTestDestinationTrait;

/**
 * Elasticsearch Destination Queue Test.
 *
 *  @group data_pipelines
 *
 *  @covers \Drupal\data_pipelines_elasticsearch\Plugin\DatasetDestination\ElasticSearchDestination
 */
class ElasticsearchDestinationQueueTest extends DatasetKernelTestBase {

  use DatasetTestProcessingTrait;
  use ElasticsearchTestDestinationTrait;

  /**
   * The data set entity.
   *
   * @var \Drupal\data_pipelines\Entity\DatasetInterface
   */
  protected DatasetInterface $dataset;

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'data_pipelines_elasticsearch',
    'data_pipelines_elasticsearch_test',
  ];

  /**
   * {@inheritdoc}
   */
  public function setUp() : void {
    parent::setUp();
    $this->initializeElasticClient();
    $this->dataset = $this->createTestDataset(['destinations' => [$this->destination]]);
  }

  /**
   * Tests index creation.
   *
   *  @covers \Drupal\data_pipelines_elasticsearch\Plugin\DatasetDestination\ElasticSearchDestination::getCurrentCount
   *  @covers \Drupal\data_pipelines_elasticsearch\Plugin\DatasetDestination\ElasticSearchDestination::getProcessingChunkSize
   *  @covers \Drupal\data_pipelines_elasticsearch\Plugin\DatasetDestination\ElasticSearchDestination::beginProcessing
   *  @covers \Drupal\data_pipelines_elasticsearch\Plugin\DatasetDestination\ElasticSearchDestination::createIndex
   *  @covers \Drupal\data_pipelines_elasticsearch\Plugin\DatasetDestination\ElasticSearchDestination::processChunk
   *  @covers \Drupal\data_pipelines_elasticsearch\Plugin\DatasetDestination\ElasticSearchDestination::endProcessing
   */
  public function testIndexing() {
    $this->assertEquals(0, $this->destinationPlugin->getCurrentCount($this->dataset));
    $queue_id = $this->dataset->getProcessingQueueId();
    $process_queue = \Drupal::queue($queue_id);
    $this->dataset->queueProcessing($process_queue);
    // The ES destination index items in bulk so only one process chunk.
    $this->assertEquals(3, $process_queue->numberOfItems());
    $this->processQueueWithWait($queue_id, $this->destination->id(), [
      ProcessingOperationEnum::BEGIN,
      ProcessingOperationEnum::PROCESS_CHUNK,
      ProcessingOperationEnum::END,
    ]);
    $this->assertTrue($this->client->indices()->exists(['index' => 'foo_' . $this->dataset->getMachineName()]));
    $this->assertEquals(2, $this->destinationPlugin->getCurrentCount($this->dataset));

    // Set unpublished.
    $this->dataset->setUnpublished();
    $this->dataset->save();
    $this->assertFalse($this->client->indices()->exists(['index' => 'foo_' . $this->dataset->getMachineName()]));

    // Set published.
    $this->dataset->setPublished();
    $this->dataset->save();
    $this->dataset->queueProcessing($process_queue);
    // The ES destination index items in bulk so only one process chunk.
    $this->assertEquals(3, $process_queue->numberOfItems());
    $this->processQueueWithWait($queue_id, $this->destination->id(), [
      ProcessingOperationEnum::BEGIN,
      ProcessingOperationEnum::PROCESS_CHUNK,
      ProcessingOperationEnum::END,
    ]);
    $this->assertTrue($this->client->indices()->exists(['index' => 'foo_' . $this->dataset->getMachineName()]));
    $this->assertEquals(2, $this->destinationPlugin->getCurrentCount($this->dataset));

    // Clean up.
    $this->dataset->delete();
    $this->assertFalse($this->client->indices()->exists(['index' => 'foo_' . $this->dataset->getMachineName()]));

  }

  /**
   * Tests retain invalid data removal.
   */
  public function testRetainInvalid(): void {
    $queue_id = $this->dataset->getProcessingQueueId();
    /** @var \Drupal\Core\Queue\QueueWorkerInterface $queue_worker */
    $process_queue = \Drupal::queue($queue_id);
    $this->dataset->queueProcessing($process_queue);
    // The elastic destination index all items at once.
    $this->assertEquals(3, $process_queue->numberOfItems());
    $this->processQueueWithWait($queue_id, $this->destination->id(), [
      ProcessingOperationEnum::BEGIN,
      ProcessingOperationEnum::PROCESS_CHUNK,
      ProcessingOperationEnum::END,
    ]);
    $this->assertEquals(2, $this->destinationPlugin->getCurrentCount($this->dataset));
    // Add some more data to the dataset including some invalid data.
    $csv = <<<CSV
should_we,firstname,lastname
Y,joe,bloggs
N,betty,bloggs
N,li,bloggs
Y,nick,bloggs
N,nora,bloggs
CSV;
    $this->dataset->csv_text->value = $csv;
    $this->dataset->save();
    $this->dataset->queueProcessing($process_queue);
    // The elastic destination index all items at once. Item 3 is invalid but
    // still we have no clean up item as we want to retain them.
    $this->assertEquals(3, $process_queue->numberOfItems());
    $this->processQueueWithWait($queue_id, $this->destination->id(), [
      ProcessingOperationEnum::BEGIN,
      ProcessingOperationEnum::PROCESS_CHUNK,
      ProcessingOperationEnum::END,
    ]);
    // Only 4 valid items are indexed.
    $this->assertEquals(4, $this->destinationPlugin->getCurrentCount($this->dataset));
    // Remove some more data to the dataset.
    $csv = <<<CSV
should_we,firstname,lastname
N,li,bloggs
Y,mike,bloggs
N,lara,bloggs
CSV;
    $this->dataset->csv_text->value = $csv;
    $this->dataset->save();
    $this->dataset->queueProcessing($process_queue);
    // The elastic destination index all items at once. Item 1 is invalid but
    // still we have no clean up item as we want to retain them.
    $this->assertEquals(3, $process_queue->numberOfItems());
    $this->processQueueWithWait($queue_id, $this->destination->id(), [
      ProcessingOperationEnum::BEGIN,
      ProcessingOperationEnum::PROCESS_CHUNK,
      ProcessingOperationEnum::END,
    ]);
    // Three older items and two new item all are indexed.
    $this->assertEquals(5, $this->destinationPlugin->getCurrentCount($this->dataset));
  }

  /**
   * Tests remove invalid data removal.
   */
  public function testRemoveInvalid(): void {
    $this->dataset->invalid_values->value = DatasetInterface::INVALID_REMOVE;
    $this->dataset->save();
    $this->assertEquals(0, $this->destinationPlugin->getCurrentCount($this->dataset));
    $queue_id = $this->dataset->getProcessingQueueId();
    /** @var \Drupal\Core\Queue\QueueWorkerInterface $queue_worker */
    $process_queue = \Drupal::queue($queue_id);
    $this->dataset->queueProcessing($process_queue);
    // The elastic destination index all items at once.
    $this->assertEquals(3, $process_queue->numberOfItems());
    $this->processQueueWithWait($queue_id, $this->destination->id(), [
      ProcessingOperationEnum::BEGIN,
      ProcessingOperationEnum::PROCESS_CHUNK,
      ProcessingOperationEnum::END,
    ]);
    $this->assertEquals(2, $this->destinationPlugin->getCurrentCount($this->dataset));
    // Add some more data to the dataset including some invalid data.
    $csv = <<<CSV
should_we,firstname,lastname
Y,joe,bloggs
N,betty,bloggs
N,li,bloggs
Y,nick,bloggs
N,nora,bloggs
CSV;
    $this->dataset->csv_text->value = $csv;
    $this->dataset->save();
    $this->dataset->queueProcessing($process_queue);
    // The elastic destination index all items at once. Item 3 is invalid.
    $this->assertEquals(4, $process_queue->numberOfItems());
    $this->processQueueWithWait($queue_id, $this->destination->id(), [
      ProcessingOperationEnum::BEGIN,
      ProcessingOperationEnum::PROCESS_CHUNK,
      ProcessingOperationEnum::PROCESS_CLEANUP,
      ProcessingOperationEnum::END,
    ]);
    $this->assertEquals(4, $this->destinationPlugin->getCurrentCount($this->dataset));
    // Remove some more data to the dataset.
    $csv = <<<CSV
should_we,firstname,lastname
N,li,bloggs
Y,mike,bloggs
N,lara,bloggs
CSV;
    $this->dataset->csv_text->value = $csv;
    $this->dataset->save();
    $this->dataset->queueProcessing($process_queue);
    // The elastic destination index all items at once. Item 1 is valid.
    $this->assertEquals(5, $process_queue->numberOfItems());
    $this->processQueueWithWait($queue_id, $this->destination->id(), [
      ProcessingOperationEnum::BEGIN,
      ProcessingOperationEnum::PROCESS_CHUNK,
      // One clean up to remove the invalid.
      ProcessingOperationEnum::PROCESS_CLEANUP,
      // Second clean up to remove the outdated.
      ProcessingOperationEnum::PROCESS_CLEANUP,
      ProcessingOperationEnum::END,
    ]);
    $this->assertEquals(2, $this->destinationPlugin->getCurrentCount($this->dataset));
  }

}
