<?php

declare(strict_types=1);

namespace Drupal\Tests\data_pipelines_elasticsearch\Traits;

use Drupal\data_pipelines\Destination\DatasetDestinationPluginInterface;
use Drupal\data_pipelines\Entity\Destination;
use Drupal\data_pipelines\Entity\DestinationInterface;
use Elasticsearch\Client;
use Elasticsearch\Common\Exceptions\NoNodesAvailableException;

/**
 * Elasticsearch Destination Trait.
 */
trait ElasticsearchTestDestinationTrait {

  /**
   * The destination entity.
   *
   * @var \Drupal\data_pipelines\Entity\DestinationInterface
   */
  protected DestinationInterface $destination;

  /**
   * The destination plugin instance.
   *
   * @var \Drupal\data_pipelines\Destination\DatasetDestinationPluginInterface
   */
  protected DatasetDestinationPluginInterface $destinationPlugin;

  /**
   * The Elasticsearch client.
   *
   * @var \Elasticsearch\Client|null
   */
  protected ?Client $client;

  /**
   * Creates a test elasticsearch destination.
   *
   * @return \Drupal\data_pipelines\Entity\DestinationInterface
   *   The test destination.
   */
  protected function createElasticsearchDestination(array $values = []): DestinationInterface {
    $destinationId = $values['destinationId'] ?? $this->randomMachineName();
    $destination = Destination::create($values + [
      'id' => $destinationId,
      'label' => 'Test destination',
      'destination' => 'elasticsearch',
      'destinationSettings' => [
        'url' => '',
        'username' => '',
        'password' => '',
        'prefix' => '',
      ],
    ]);
    $destination->save();
    return $destination;
  }

  /**
   * Initializes Elastic client.
   *
   *   @covers \Drupal\data_pipelines_elasticsearch\Plugin\DatasetDestination\ElasticSearchDestination::getClient
   */
  protected function initializeElasticClient():void {
    $this->destination = $this->createElasticsearchDestination([
      'destinationSettings' => [
        'url' => getenv('ELASTICSEARCH_URL'),
        'username' => 'elastic',
        'password' => 'changeme',
        'prefix' => 'foo_',
      ],
    ]);
    $this->destinationPlugin = $this->destination->getDestinationPlugin();
    $this->client = $this->destinationPlugin->getClient();
    $this->assertInstanceOf(Client::class, $this->client);
    try {
      $this->client->ping();
    }
    catch (NoNodesAvailableException $exception) {
      $this->markTestSkipped('The test is skipped as no alive nodes found in cluster');
    }

  }

}
